﻿#pragma once

#ifdef _DEBUG
#include <iostream>
#include <unordered_map>
#include <chrono>

namespace cig::Memory::Debug
{
	namespace PrettyDurationDump
	{
		inline std::string DumpDurationPretty(const std::chrono::nanoseconds& duration)
		{
			using timetype = unsigned long long;
			struct
			{
				bool hours = false;
				bool minutes = false;
				bool seconds = false;
				bool milliseconds = false;
				bool microseconds = false;
			} m_TimeFlag;

			timetype precice_nanoseconds = duration.count();
			timetype nanoseconds = precice_nanoseconds % 1000;

			timetype precice_microseconds = precice_nanoseconds / 1000;
			timetype microseconds = precice_microseconds % 1000;
			if (microseconds > 0) m_TimeFlag.microseconds = true;

			timetype precice_milliseconds = precice_microseconds / 1000;
			timetype milliseconds = precice_milliseconds % 1000;
			if (milliseconds > 0) m_TimeFlag.milliseconds = true;

			timetype precice_seconds = precice_milliseconds / 1000;
			timetype seconds = precice_seconds % 60;
			if (seconds > 0) m_TimeFlag.seconds = true;

			timetype precice_minutes = precice_seconds / 60;
			timetype minutes = precice_minutes % 60;
			if (minutes > 0) m_TimeFlag.minutes = true;

			timetype precice_hours = precice_minutes / 60;
			timetype hours = precice_hours;
			if (hours > 0) m_TimeFlag.hours = true;

			std::stringstream ss;

			if (m_TimeFlag.hours)
				ss << hours << "h ";
			if (m_TimeFlag.minutes)
				ss << minutes << "m ";
			if (m_TimeFlag.seconds)
				ss << seconds << "s ";
			if (m_TimeFlag.milliseconds)
				ss << milliseconds << "ms ";
			if (m_TimeFlag.microseconds)
				ss << microseconds << "us ";
			ss << nanoseconds << "ns";

			return ss.str();
		}
	}

	struct ptr_info
	{
		size_t count, bytes;
		const char* var;
		const char* file;
		const char* func;
		int line;
	};

	inline std::unordered_map<void*, ptr_info> allocations;

	template<typename T>
	void log_allocation(const T* ptr, const size_t size, const char* var = "", const char* file = "", const char* func = "", int line = 0)
	{
		allocations[(void*)ptr] = {
			size,
			sizeof(T),
			var,
			file,
			func,
			line
		};
	}

	template<typename T>
	void log_deallocator(const T* ptr)
	{
		allocations.erase(allocations.find((void*)ptr));
	}

	template<typename T>
	void log_reallocator(const T* ptr, const T* newptr, const size_t to, const char* var = "", const char* file = "", const char* func = "", int line = 0)
	{
		log_allocation(newptr, to, var, file, func, line);
	}

	template<typename T>
	void log_dupallocator(const T* ptr, const size_t size, const char* var = "", const char* file = "", const char* func = "", int line = 0)
	{
		log_allocation(ptr, size, var, file, func, line);
	}

	inline void log_info_dump(const char* file, const char* func, int line)
	{
		static std::chrono::high_resolution_clock::time_point s_LastDump;
		static bool s_FirstDump = true;
		if (s_FirstDump)
		{
			s_LastDump = std::chrono::high_resolution_clock::now();
			s_FirstDump = false;
			return;
		}

		std::cout << "\n------------------------------------ Allocator Debug Info ------------------------------------" << std::endl;
		if (file)
			std::cout << "\tFile: \t\t" << file << std::endl;
		if (func)
			std::cout << "\tFunction: \t" << func << std::endl;
		if (line)
			std::cout << "\tLine: \t\t" << line << std::endl;
		// time since last dump

		auto now = std::chrono::high_resolution_clock::now();
		const auto diff = std::chrono::duration_cast<std::chrono::nanoseconds>(now - s_LastDump);

		std::cout << "\tLast dump:\t" << PrettyDurationDump::DumpDurationPretty(diff) << " ago" << std::endl;
		s_LastDump = std::chrono::high_resolution_clock::now();

		std::cout << "\tAllocations: " << allocations.size() << std::endl;
		for (auto& [ptr, info] : allocations)
		{
			std::cout << "Allocation: " << ptr
				<< "\n\tSize: " << info.count
				<< "\n\tBytes: " << info.bytes * info.count
				<< "\n\tFile: " << info.file
				<< "\n\tFunction: " << info.func
				<< "\n\tLine: " << info.line
				<< "\n\tVariable: " << info.var
				<< std::endl;
		}
		std::cout << "----------------------------------------------------------------------------------------------" << std::endl;
	}
}

#define MEMORY_ALLOCATOR_DEBUG(ptr, size) cig::Memory::Debug::log_allocation(ptr, size, #ptr, __FILE__, __FUNCTION__, __LINE__)
#define MEMORY_DEALLOCATOR_DEBUG(ptr) cig::Memory::Debug::log_deallocator(ptr)
#define MEMORY_REALLOCATOR_DEBUG(ptr, newptr, to) cig::Memory::Debug::log_reallocator(ptr, newptr, to, #newptr, __FILE__, __FUNCTION__, __LINE__)
#define MEMORY_DUPALLOCATOR_DEBUG(ptr, size) cig::Memory::Debug::log_dupallocator(ptr, size, #ptr, __FILE__, __FUNCTION__, __LINE__)
#define MEMORY_INFO_DUMP() cig::Memory::Debug::log_info_dump(__FILE__, __FUNCTION__, __LINE__)
#else
#define MEMORY_ALLOCATOR_DEBUG(ptr, size)
#define MEMORY_DEALLOCATOR_DEBUG(ptr)
#define MEMORY_REALLOCATOR_DEBUG(ptr, from, to)
#define MEMORY_DUPALLOCATOR_DEBUG(ptr, size)
#endif